import json, sys, aiohttp
from .requests import Manager
from .addons import Utils, Signal, Debugger, lunarlist, _, Database, Cooldown, inspect, os
from ..errors import ConnectionClosedError
from ..basetypes import Message, Channel, User, Interaction, Context, Reaction, Guild, Role, Snowflake, Member
from ..commands import SlashCommand
from ..ui import Button, View, SelectMenu, SelectMenuOption, Modal, TextInput
import time, threading, traceback, zlib

debug = Debugger(
    method=1
)

debug.disable()

GUILDS = 1
GUILD_MEMBERS = 2
GUILD_MODERATION = 4
EMOJIS_AND_STICKERS = 8
GUILD_EMOJIS_AND_STICKERS = EMOJIS_AND_STICKERS
GUILD_EMOJIS = EMOJIS_AND_STICKERS
GUILD_STICKERS = EMOJIS_AND_STICKERS
GUILD_INTEGRATIONS = 16
GUILD_WEBHOOKS = 32
GUILD_INVITES = 64
GUILD_VOICE_STATES = 128
GUILD_PRESENCES = 256
GUILD_MESSAGES = 512
GUILD_REACTIONS = 1024
GUILD_MESSAGE_REACTIONS = 1024
GUILD_TYPING = 2048
GUILD_MESSAGE_TYPING = 2048
DIRECT_MESSAGES = 4096
DIRECT_MESSAGE_REACTIONS = 8192
DIRECT_MESSAGE_TYPING = 16384
MESSAGE_CONTENT = 32768
MESSAGES_CONTENT = 32768
GUILD_SCHEDULED_EVENT = 65536
AUTOMOD_CONFIGURATION = 1048576
AUTO_MOD_CONFIGURATION = 1048576
AUTO_MODERATION_CONFIGURATION = 1048576
AUTOMOD_EXECUTION = 2097152
AUTO_MOD_EXECUTION = 2097152
AUTO_MODERATION_EXECUTION = 2097152
ALL_INTENTS = 3276799
ALL = 3276799

# All opcode IDs

DISPATCH = 0
HEARTBEAT = 1
IDENTIFY = 2
PRESENCE_UPDATE = 3
VOICE_STATE_UPDATE = 4
RESUME = 6
RECONNECT = 7
REQUEST_GUILD_MEMBERS = 8
INVALID_SESSION = 9
HELLO = 10
HEARTBEAT_ACK = 11

# All opcode names

OPCODES = [
    'DISPATCH',
    'HEARTBEAT',
    'IDENTIFY',
    'PRESENCE_UPDATE',
    'VOICE_STATE_UPDATE',
    None,
    'RESUME',
    'RECONNECT',
    'REQUEST_GUILD_MEMBERS',
    'INVALID_SESSION',
    'HELLO',
    'HEARTBEAT_ACK'
]

# Relationship Types

NO_RELATIONSHIP = 0
FRIEND = 1
BLOCKED = 2
PENDING_INCOMING = 3
PENDING_OUTGOING = 4
IMPLICIT = 5

class Relationship:
    type = -1
    name = 'Base'
    
class NoRelationship(Relationship):
    type = 0
    name = 'Null'
    
class Friend(Relationship):
    type = 1
    name = 'Friend'
    
class Blocked(Relationship):
    type = 2
    name = 'Blocked'
    
class PendingIncoming(Relationship):
    type = 3
    name = 'Pending Incoming'
    
class PendingOutgoing(Relationship):
    type = 4
    name = 'Pending Outgoing'
    
class Implicit(Relationship):
    type = 5
    name = 'Implicit'
    
RELATIONSHIP_TYPES = {relationship.type: relationship.name for relationship in (NoRelationship, Friend, Blocked, PendingIncoming, PendingOutgoing, Implicit)}

class GatewaySession:
    '''
    Represents a gateway's session data.
    '''
    
    def __init__(self, gateway, data: dict, /):
        
        self.gateway: Gateway = gateway
        self.raw: dict = data
        self.type: str = data.get('session_type')
        self.id: str = data.get('session_id')
        self.resume_url: str = data.get('resume_gateway_url')
        self.client_info: dict = data.get('client_info')
        self.status: str = data.get('status')
        self.active: bool = data.get('active', False)
        self.created_at: int = data.get('created_at')
        self.activities: list[dict] = data.get('activities', [])
        
        if self.type is None:
            self.type: str = data.get('type')
            
        self.resumer: GatewayResume = GatewayResume(self)
        
    def __repr__(self):
        return f'GatewaySession(gateway={self.gateway}, data={self.raw})'
        

class GatewayResume:
    '''
    Represents data used for easily resuming the gateway connection/session.
    '''
    
    def __init__(self, session: GatewaySession, /):
        
        self.session: GatewaySession = session
        self.url: str = self.session.resume_url
        
    @property
    def gateway(self):
        return self.session.gateway
    
    @property
    def token(self):
        return self.gateway.token
        
    @property
    def sequence(self):
        return self.gateway.__sequence__
    
    def __repr__(self):
        return f'GatewayResume(session={self.session})'
    
    async def resume(self):
        
        if self.url is None:
            return None
        
        token = self.token
        sequence = self.sequence
        
        oldGateway = self.gateway
        
        self.session.gateway = None
        
        newGateway = Gateway(oldGateway.bot, oldGateway)
        newGateway.setToken(token)
        
        oldGateway.bot.__gateway__ = newGateway
        await oldGateway.stopSignal.call()
        
        await newGateway.start(
            runAfterThread = oldGateway.runAfterThread,
            inThread = True
        )
        
        self.session.gateway = newGateway
        
        await self.gateway.send(
            opcode = 6,
            token = self.token,
            session_id = self.session.id,
            seq = sequence
        )
        
        self.session = None
        del oldGateway


class Intent:
    
    def __init__(self, intents, id):
        
        self.id: int = id
        self.intents: Intents = intents
    
    def enable(self):
        
        if self.id is None:
            return None
        
        self.intents.append(self.id)
        
    def disable(self):
        
        if self.id is None:
            return None
        
        self.intents.remove(self.id)
        
    def __int__(self):
        
        return self.id if self.id is not None else 0
    
    def __add__(self, other: int):
        
        if type(other) is not int:
            
            try:
                other = int(other)
                
            except:
                other = 0
                
        return int(self) + other

class Intents:
    
    def __init__(self, bot):
        
        from ..bot import Bot
        
        self.bot: Bot = bot
        
        self.__intents__: list[Intents] = []
        
        self.guilds = Intent(self, GUILDS)
        self.guildMembers = Intent(self, GUILD_MEMBERS)
        self.guildModeration = Intent(self, GUILD_MODERATION)
        self.emojisAndStickers = Intent(self, EMOJIS_AND_STICKERS)
        self.guildIntegrations = Intent(self, GUILD_INTEGRATIONS)
        self.guildWebhooks = Intent(self, GUILD_WEBHOOKS)
        self.guildInvites = Intent(self, GUILD_INVITES)
        self.guildVoiceStates = Intent(self, GUILD_VOICE_STATES)
        self.guildPresences = Intent(self, GUILD_PRESENCES)
        self.guildMessages = Intent(self, GUILD_MESSAGES)
        self.guildReactions = Intent(self, GUILD_REACTIONS)
        
    def append(self, intent: Intent):
        
        if type(intent) is not intent:
            return None
        
        if intent in self.__intents__:
            return None
        
        self.__intents__.append(intent)
        
    def remove(self, intent: Intent):
        
        if type(intent) is not intent:
            return None
        
        if intent not in self.__intents__:
            return None
        
        self.__intents__.remove(intent)
        
    def __iter__(self):
        
        return self.__intents__
    
    def toInteger(self) -> int:
        
        value = 0
        
        for x in self:
            value += x
            
        return value
    
    def __int__(self):
        
        return self.toInteger()
        

class Gateway:
    
    def __init__(self, bot, previousGateway = None):
        
        from ..bot import Bot, Loop, Task
        
        previousGateway: Gateway = previousGateway
        
        hasPrevious: bool = previousGateway is not None
        isFirst: bool = previousGateway is None
        
        self.ws: aiohttp.ClientWebSocketResponse = None
        self.bot: Bot = bot
        self.asyncio = self.bot.asyncio
        self.token: str = None
        self.gatewayUrl: str = 'wss://gateway.discord.gg/?v=9&encoding=json'
        self.isBot: bool = True
        self.origin: str = self.bot.origin
        
        self.db = Database(
            origin = self.origin
        )
        
        global asyncio
        asyncio = self.asyncio
        
        if hasPrevious: # Restore previous attributes
            
            self.httpSession: aiohttp.ClientSession = previousGateway.httpSession
            self.manager: Manager = previousGateway.manager
            
            self.relationships: lunarlist[dict] = previousGateway.relationships
            self.users: lunarlist[User] = previousGateway.users
            self.channels: lunarlist[Channel] = previousGateway.channels
            self.guilds: lunarlist[Guild] = previousGateway.guilds
            
            self.onStartFunctions: list = previousGateway.onStartFunctions
            self.onMessageFunctions: list = previousGateway.onMessageFunctions
            self.channelUpdateFunctions: list = previousGateway.channelUpdateFunctions
            self.channelDeleteFunctions: list = previousGateway.channelDeleteFunctions
            self.channelCreateFunctions: list = previousGateway.channelCreateFunctions
            self.typingStartedFunctions: list = previousGateway.typingStartedFunctions
            self.typingStoppedFunctions: list = previousGateway.typingStoppedFunctions
            self.presenceUpdateFunctions: list = previousGateway.presenceUpdateFunctions
            self.messageUpdatedFunctions: list = previousGateway.messageUpdatedFunctions
            self.messageDeletedFunctions: list = previousGateway.messageDeletedFunctions
            self.reactionAddFunctions: list = previousGateway.reactionAddFunctions
            self.reactionRemoveFunctions: list = previousGateway.reactionRemoveFunctions
            self.messagePinnedFunctions: list = previousGateway.messagePinnedFunctions
            self.memberJoinFunctions: list = previousGateway.memberJoinFunctions
            self.memberLeaveFunctions: list = previousGateway.memberLeaveFunctions
            self.memberUpdateFunctions: list = previousGateway.memberUpdateFunctions
            
            self.typingTimes: dict[tuple[int], float] = previousGateway.typingTimes
            self.eventSignals: dict[str, list[Signal]] = previousGateway.eventSignals

            self.views: list[View] = previousGateway.views
            self.modals: list[Modal] = previousGateway.modals
            self.loops: list[Loop] = previousGateway.loops
            self.tasks: list[Task] = previousGateway.tasks
            
            self.isBot: bool = True
            
        else: # Create new attributes
            
            self.httpSession: aiohttp.ClientSession = aiohttp.ClientSession(connector=aiohttp.TCPConnector(limit=200))
            self.manager: Manager = Manager(self.token, self.httpSession, self.isBot, self)
            
            self.relationships: lunarlist[dict] = lunarlist()
            self.users: lunarlist[User] = lunarlist()
            self.channels: lunarlist[Channel] = lunarlist()
            self.guilds: lunarlist[Guild] = lunarlist()
        
            self.onStartFunctions: list = []
            self.onMessageFunctions: list = []
            self.channelUpdateFunctions: list = []
            self.channelCreateFunctions: list = []
            self.channelDeleteFunctions: list = []
            self.typingStartedFunctions: list = []
            self.typingStoppedFunctions: list = []
            self.presenceUpdateFunctions: list = []
            self.messageUpdatedFunctions: list = []
            self.messageDeletedFunctions: list = []
            self.reactionAddFunctions: list = []
            self.reactionRemoveFunctions: list = []
            self.messagePinnedFunctions: list = []
            self.memberLeaveFunctions: list = []
            self.memberJoinFunctions: list = []
            self.memberUpdateFunctions: list = []
            
            self.typingTimes: dict[tuple[int], float] = {}
            self.eventSignals: dict[str, list[Signal]] = {}

            self.views: list[View] = []
            self.modals: list[Modal] = []
            self.loops: list[Loop] = []
            self.tasks: list[Task] = []
            
            for x in ('READY', 'MESSAGE_CREATE', 'INTERACTION_CREATE'):
                
                self.eventSignals[x] = []
            
        if self.httpSession.closed:
            self.httpSession: aiohttp.ClientSession = aiohttp.ClientSession(connector=aiohttp.TCPConnector(limit=200))
        
        self.slashCommands: list[SlashCommand] = []
        
        self.__decompress__: zlib._Decompress = zlib.decompressobj()
        
        self.__sequence__: int = None
        self.heartbeatInterval: float = None
        self.last_heartbeat_ack: float = None
        self.last_heartbeat: float = None
        self.latency: float = None
        
        self.closed: bool = False
        self.resumed: bool = hasPrevious
        self.forceFullyTerminated: bool = False # If True, bot will not try to reconnect.
        self.reconnect: Signal = Signal()
        self.reconnect.connect(self.resume)
        
        try:
            self.eventLoop = asyncio.get_event_loop() # The eventloop used for asyncio stuff
            
        except:
            self.eventLoop = None
        
        self.isReady: bool = False
        
    @property
    def preToken(self):
        return 'Bot ' if self.isBot else ''
    
    @property
    def fullToken(self):
        return self.preToken + self.token
    
    @property
    def messages(self):
        '''
        All the messages in the bot's cache in one generator.
        '''
        
        for channel in self.channels:
            yield from channel.messages

    @property
    def roles(self):
        """
        All roles in the bot's cache, from all the different guilds the bot is in.
        """

        for guild in self.guilds:
            yield from guild.roles

    @property
    def members(self):
        """
        All members in the bot's cache, from all the guilds the bot is in.
        """

        for guild in self.guilds:
            yield from guild.members
            
    @property
    def prefixes(self) -> list[str]:
        return self.bot.prefixes
    
    @property
    def uptime(self) -> float:
        return time.time() - self.startTime
    
    async def resume(self):
        
        if not hasattr(self, 'session'):
            return None
        
        if self.forceFullyTerminated:
            return None
        
        session = self.session
        
        if session is None:
            return None
        
        resumer = session.resumer
        
        if resumer is None:
            return None
        
        await resumer.resume()
        
    def getSession(self):
        
        '''
        Gets the current `ClientSession` or creates a new one if needed.
        '''
        
        try:
            if self.httpSession is not None and not self.httpSession.closed:
                return self.httpSession
            
        except:
            
            ...
            
        return self.setSession()
        
    def setSession(self):
        
        '''
        Creates a new `ClientSession`, updates the current one and returns it.
        '''
        
        self.httpSession = aiohttp.ClientSession(connector=aiohttp.TCPConnector(limit=200))
        self.manager.session = self.httpSession
        return self.httpSession
    
    async def getWs(self):
        
        '''
        Gets the current (WebSocket) or creates a new one if needed.
        '''
        
        if not hasattr(self, 'ws') or self.ws is None or self.ws.closed:
            return await self.setWs()
        
        return self.ws
    
    async def setWs(self):
        
        '''
        Creates a new `ClientWebSocketResponse` (WebSocket), returning it and updating the current one.
        '''
        
        try:
            self.ws = await self.getSession()._ws_connect(self.gatewayUrl, timeout=30, autoclose=True)
        
        except:
            self.ws = await self.setSession()._ws_connect(self.gatewayUrl, timeout=30, autoclose=True)
            
        return self.ws
    
    def __updateSavedViews__(self):

        views: list[dict] = self.db.loadViews()

        for view in views:

            messages = view.get("msgs")
            nm = view.get("nm")
            last = view.get("lst")

            for model in self.views:
                if model.name == nm:
                    model.messages = messages
                    model.last = last

        
    def setToken(self, token: str):
        self.token = token
        self.manager.generateHeaders(self.fullToken)
        
    def toJson(self, payload):
        return json.dumps(payload)
        
    async def send(self, opcode: int, replacement = None, **data):
        
        if self.closed:
            return None
        
        if hasattr(self, 'ws') and self.ws:
            
            if data == {}:
                data = replacement
                
            payload = {
                'op': opcode,
                'd': data
            }
            
            payload = self.toJson(payload)
            
            try:
                
                await self.ws.send_str(
                    data = payload
                )
                
                return payload
                
            except:
                
                #print("Failed to send payload", payload)
                self.closed = True
                return False
                #raise ConnectionClosedError()
            
    async def receive(self, timeout: float = None):
        
        if self.closed:
            return None
        
        try:
            
            message = await self.ws.receive(timeout=timeout)
            payload = message.data
            
        except:
            self.closed = True
            return 'CLOSED'
            #raise ConnectionClosedError()
        
        if isinstance(payload, bytes):
            try:
                payload = payload.decode('utf-8')
            except:
                buffer = bytearray()
                buffer.extend(payload)
                payload = self.__decompress__.decompress(buffer)
                payload = payload.decode('utf-8')
                
        '''debug.log(
            'Received data! Raw form: %p',
            payload
        )'''
        
        if payload:

            try:
                return json.loads(payload)
            except:
                return payload
        
    async def identify(self):
        
        data = await self.send(
            
            opcode = 2,
            token = self.token,
            properties = {
                "os": sys.platform,
                "browser": "lunarcord",
                "device": "lunarcord"
            },
            
            intents = ALL,
            presence = self.bot.presence,
            large_threshold = 250
            
        )
        
    async def sendHeartbeat(self):

        #print(self.__sequence__)
        
        result = await self.send(
            
            opcode = 1,
            replacement = self.__sequence__
            
        )

        if result is not False:

            ... #print("Heartbeat sent! Gateway:", self)
        
        
    async def changeStatus(self, activities, status: str = 'online'):
        
        await self.send(
            
            opcode = 3,
            since = 0,
            activities = activities,
            status = status,
            afk = False,
            client_status = {'mobile': 'online'}
            
        )
        
    def doTask(self, *coroutines, signal: Signal = None):
        
        '''
        Start running one or more coroutines as tasks in the event loop, calling `signal` when one is finished. The signal args will be `coroutine` (the coroutine that was completed) and `result` (the coroutine's result).
        '''
        
        coroutines = list(coroutines)
        tasks: list = list()
        
        def finished(task):
            
            nonlocal tasks, coroutines, signal
            
            try:
                result = task.result()
                
            except:
                result = None
                
            try:
                idx = tasks.index(task)
            
            except:
                idx = 0
                
            try:
                coro = coroutines[idx]
                
            except:
                
                try:
                    coro = coroutines[0]
                    
                except:
                    coro = None
            
            # tasks = All scheduled tasks.
            # coroutines = The coroutines given in doTask args.
            # task = Task that was finished.
            # coro = The coroutine which was wrapped in this task.
            # signal = Signal to emit.
            # result = The task's results.
            
            if signal is not None:
                signal.callSync(coro, result)
                
        try: # Method 1
            
            tasks = []
                
            fut = asyncio.gather(*coroutines)
            
            async def gather():
                nonlocal fut
                await fut
                
            coro = gather()
            
            task = self.eventLoop.create_task(coro)
            #task.add_done_callback(finished)
            tasks.append(task)
                
            return tasks
        
        except: # Method 2
            
            tasks = []
            
            for coro in coroutines:
                
                tasks.append(
                    self.eventLoop.create_task(coro)
                )
                
            return tasks
    
    def finishTasks(self, *coroutines):
        
        '''
        Similarly to `doTask`, executes one or more coroutines. The differences are that this is done in a blocking manner and the results are returned, unlike with `Gateway.doTask`.
        '''
        
        def onFinish(coro, result):
            
            nonlocal pending, finished, returns
            
            if len(pending) <= 0:
                finished = True
                
            try:
                pending.remove(coro)
                
            except:
                ...
                
            returns.append(result)
        
        finished = False
        returns = []
        
        coroutines = list(coroutines)
        pending = coroutines.copy()
        signal = Signal(object, object)
        signal.connect(onFinish)
        
        self.doTask(*coroutines, signal=signal)
        
        while not finished: # Block until finished
            continue # ...
        
        return returns
    
    def finishTask(self, coroutine):
        
        '''
        Same as using `Gateway.finishTasks(coroutine)[0]`.
        '''
        
        coroutines: list = [coroutine]
        
        results = self.finishTasks(*coroutines)
        return results[0]
        
    def runInThread(self, function, *args, signal: Signal = None, name: str = None, **kwargs):
        
        '''
        Run a normal function or a coroutine function inside a Thread, calling `signal` upon finished.
        '''
        
        args = list(args)
            
        def run():

            nonlocal self, function, args, kwargs, signal
                        
            result = function(*args, **kwargs)
            
            if asyncio.iscoroutine(result):
                result = self.doTask(result)

            if signal is not None:
                signal.callSync(result)
            
        thread = threading.Thread(target=run, name=name)
        thread.start()
        return thread
        
        
    async def start(self, runAfterThread: threading.Thread, inThread: bool = False):
        
        try:
            self.eventLoop = asyncio.get_event_loop()
        except:
            self.eventLoop = asyncio.new_event_loop()
        
        if not self.eventLoop.is_running():
            self.eventLoop.run_forever()
        
        self.stopSignal = Signal() # A signal to stop all threads.
        self.stopSignal.connect(self.httpSession.close)
        await self.setWs()
        await self.bot.__authenticate__()
        self.runAfterThread = runAfterThread
        
        try:
        
            if not runAfterThread.is_alive():
                runAfterThread.start()
                
        except:
            
            ...
            
        if inThread:
            
            self.runInThread(self.gatewayLoop, self.ws, self.stopSignal)
            
        else:
            
            await self.gatewayLoop(self.ws, self.stopSignal)
        
    def stop(self, code: int = None):
        
        for task in self.tasks + self.loops:
            
            try:
                task.stop()
                
            except:
                continue
        
        self.forceFullyTerminated = True
            
        coros = []
        
        if self.ws is not None:
            
            coros.append(
                
                self.ws.close(
                    
                    *
                    [x for x in [code]
                    if x is not None]
                    
                )
                
            )
            
        if self.httpSession is not None:
            
            coros.append(
                self.httpSession.close()
            )
            
        self.doTask(*coros)
            
        if hasattr(self, 'stopSignal') and self.stopSignal:
            self.stopSignal.callSync()
            
        self.closed = True
        
    
    async def disconnect(self, reconnect: bool = False):
        
        await self.stop()
        
        if reconnect:
            await self.reconnect.call()
            
        else:
            self.forceFullyTerminated = True
        
        
    def heartbeatLoop(self, stop: Signal):
        
        sending = True
        
        def stopSending():
            
            nonlocal sending
            sending = False
            
        stop.connect(stopSending)
        secs = self.heartbeatInterval / 1000
        
        while sending:
            
            if self.last_heartbeat_ack is None or time.time() - self.last_heartbeat_ack > self.heartbeatInterval / 1000:

                tick = 0

                if self.last_heartbeat_ack:
                    tick = time.time() - self.last_heartbeat_ack
                
                self.last_heartbeat = time.time()
                coro = self.sendHeartbeat()
                asyncio.run_coroutine_threadsafe(coro, self.eventLoop)
                time.sleep(secs)
                
                
    def heartbeat(self, interval: float, stop: Signal):
        '''
        Starts the `heartbeatLoop`, which sends heartbeats to Discord every `interval` seconds.
        
        Parameters
        ----------
        interval: `float`
            The time to wait before sending a heartbeat, received from `helloMessage`.
        '''
        
        self.heartbeatInterval = interval
        return self.runInThread(self.heartbeatLoop, stop)
    
    async def gatewayLoop(self, ws: aiohttp.ClientWebSocketResponse, stop: Signal, /):
        
        if self.token:
            
            self.startTime = time.time()
            helloMessage = await self.receive()
            
            if helloMessage == 'CLOSED':
                return None
            
            interv = helloMessage.get('d').get('heartbeat_interval')
            debug.log('Hello Message Received')

            self.heartbeat(interval=interv, stop=stop)

            if not self.resumed:
                
                await self.identify()
            
            receiving = True
            
            def stopCallback():
                nonlocal receiving
                receiving = False
                
            stop.connect(stopCallback)
            
            while receiving:
                
                try:

                    event = await self.receive()

                    if event == 'CLOSED':
                        stop.callSync()
                        event = None

                    if type(event) is int:
                        
                        if event != 1000:
                            #print("Closed with error code", event)
                            pass

                        receiving = False
                        event = None

                    self.last = event

                except:

                    #print("Failed to receive! Using traceback")
                    #traceback.print_exc()
                    event = None
                    #receiving = False
                    continue # This would terminate the connection, but now we just skip the iteration.
                
                if event is not None:
                    await self.__handleEvent__(event)
                    
            if not self.forceFullyTerminated:
                
                if not self.resumed:
                    
                    await self.reconnect.call()
                
    async def waitFor(self, event: str, timeout: float = None):
        
        '''
        Wait until the given event is dispatched. If `timeout` seconds have passed and the `event` still has not been dispatched, quits the waiting.
        '''
        
        eventName = event.strip().replace(' ', '_').upper()
        
        if eventName not in self.eventSignals:
            
            return None
        
        signal: Signal = Signal(object)
        
        self.eventSignals[eventName].append(
            signal
        )
        
        return await signal.wait(self.eventLoop, timeout)
                    
    async def __handleEvent__(self, payload: dict):
        
        name: str = payload.get('t')
        sequence: int = payload.get('s')
        opcode: int = payload.get('op')
        data: dict = payload.get('d')
        
        '''debug.log(
            'New Event Received! Operation Code: %p, Operation Name:  %p, Dispatch Name: %p, Resume Sequence: %p',
            opcode, OPCODES[opcode], name, sequence
        )'''
        
        if opcode == HEARTBEAT_ACK:

            self.last_heartbeat_ack = time.time()
            self.latency = self.last_heartbeat_ack - self.last_heartbeat
            
        elif opcode == DISPATCH:
                
            if sequence is not None:
                self.__sequence__ = sequence
                
            '''debug.log(
                'New dispatch! Name: %p, Payload: %p',
                name, data
            )'''
            
            if name == 'READY':
                handler = self.__ready__
                
            elif name == 'INTERACTION_CREATE':
                handler = self.__handleInteraction__
                
            elif name == 'MESSAGE_CREATE':
                handler = self.__handleMessage__
                
            elif name == 'PRESENCE_UPDATE':
                handler = self.__handlePresenceUpdate__
                
            elif name == 'CHANNEL_UPDATE':
                handler = self.__handleChannelUpdate__
                
            elif name == 'TYPING_START':
                handler = self.__handleTyping__
                
            elif name == 'GUILD_CREATE':
                handler = self.__newGuild__
                
            elif name == 'MESSAGE_UPDATE':
                handler = self.__handleMessageUpdate__
                
            elif name == 'MESSAGE_DELETE':
                handler = self.__handleMessageDelete__
                
            elif name == 'MESSAGE_REACTION_ADD':
                handler = self.__handleMessageReaction__

            elif name == 'MESSAGE_REACTION_REMOVE':
                handler = self.__handleReactionRemove__
                
            elif name == 'CHANNEL_PINS_UPDATE':
                handler = self.__handlePinsUpdate__
                
            elif name == 'GUILD_AUDIT_LOG_ENTRY_CREATE':
                handler = self.__auditLogEntryCreate__
                
            elif name == 'SESSIONS_REPLACE':
                handler = self.__sessionsReplace__
                
            elif name == 'CHANNEL_PINS_ACK':
                handler = self.__channelPinsAck__
                
            elif name == 'RESUMED':
                handler = self.__resumed__
                
            else:
                
                split = name.split("_")
                fixed = []

                for x in split:
                    x = x.lower()
                    x = list(x)
                    x[0] = x[0].upper()
                    x = "".join(x)
                    fixed.append(x)

                fixed = "".join(fixed)
                handlername = f"__handle{fixed}__"

                try:
                    handler = getattr(self, handlername)

                except:

                    handler = None

                    debug.log(
                        'Unhandled Dispatch! Name: %p',
                        name,
                        force=False
                    )

            if handler is None:

                return
                
            returns = await handler(data)
            signals = self.eventSignals.get(name)
            
            if type(returns) is not tuple:
                
                returns = (returns,)
            
            if signals:
                
                coros = [
                    signal.call(*returns) for signal in signals
                ]
                
                await asyncio.gather(*coros)
                
        else:
            
            debug.log(
                f'Unhandled Event! Opcode: %p, Name: %p',
                opcode, OPCODES[opcode]
            )
        
    async def __handleMessage__(self, data: dict):
        
        msg = await Message.__create__(
            bot=self.bot,
            data=data
        )
        
        events = [event(msg) for event in self.onMessageFunctions]
        await asyncio.gather(*events)
        
        #await self.__handleTypingStop__(msg.author, msg.channel)
        
        isCommand = False

        if (not self.prefixes) and (self.bot.textCommands):

            self.bot.prefixes.append("!")
            
        for prefix in self.prefixes:
            
            if msg.content.lower().startswith(prefix.lower()):
                
                isCommand = True
                prefixUsed = prefix
                break
                
        if isCommand:
            
            content: str = Utils.removeFirst(prefixUsed, msg.content)
            
            content = ''.join(content).strip()
            
            called = None
            
            for command in self.bot.textCommands:
                
                names = command.namesLower
                
                for name in names:
                
                    if content.lower().startswith(name):
                        
                        content: str = Utils.removeFirst(name, content)
                        content: str = content.strip()
                        args: list = content.split(' ')
                        
                        ctx = Context(
                            bot = self.bot,
                            message = msg,
                            triggerer = prefixUsed + name,
                            prefix = prefixUsed,
                            command = command
                        )
                        
                        called = command
                        break
                        
            if called:
                await called(ctx, *args)
                #await asyncio.gather(called(ctx, *args))
                
        return msg
            
    async def __handleInteraction__(self, data: dict):
        
        interaction = Interaction(
            bot=self.bot,
            data=data
        )
        
        await interaction.__proc__()
        
        if interaction.type == 2:
            
            await self.__handleApplicationCommand__(interaction)
            
        elif interaction.type == 3:
            
            await self.__handleComponentInteraction__(interaction)
            
        elif interaction.type == 5:
            
            await self.__handleModalInteraction__(interaction)
            
        return interaction
        
    async def __handleApplicationCommand__(self, interaction: Interaction):
        
        target = None
        coros = []
        
        for slashCommand in self.slashCommands:
            if str(slashCommand.id) == str(interaction.commandID):
                target = slashCommand
                break
            
        if target is not None:
            
            interaction.command = slashCommand

            cdtimer = interaction.author.getCooldown(
                name=slashCommand.name,
                type="slashcommand"
            )

            interaction.cooldown = Cooldown.construct(cdtimer)
            coros.append(slashCommand(interaction, *interaction.args))
        
        await asyncio.gather(*coros)
        
    async def __handleComponentInteraction__(self, interaction: Interaction):
        
        if interaction.componentType == View.BUTTON:
            
            await self.__handleButtonInteraction__(interaction)
            
        elif interaction.componentType == View.STRING_SELECT:
            
            await self.__handleSelectMenuInteraction__(interaction)
            
    async def __handleButtonInteraction__(self, interaction: Interaction):
        
        target = None
        
        for view in self.views:

            if interaction.message.id not in view.messageIDs:
                continue
            
            button: Button = view.getButton(
                
                customID = interaction.customID
                
            )
            
            if button:
                
                target = button
                break
            
        if target:

            if True:
            
                await interaction.ack()
                await target(interaction)
                button.view.updateLast()

            else:

                target.view.delete()
            
    async def __handleSelectMenuInteraction__(self, interaction: Interaction):
        
        target = None
        
        for view in self.views:

            if interaction.message.id not in view.messageIDs:
                continue
            
            select: SelectMenu = view.getSelectMenu(
                
                customID = interaction.customID
                
            )
            
            if select:
                
                target = select
                break
                
        if target:
            
            for idx, option in enumerate(interaction.values.copy()):
                
                for targetOption in target.options:
                    
                    if targetOption.value == option:
                        
                        interaction.values[idx] = targetOption
            
            coros = [interaction.ack(), target(interaction)]
            
            for clicked in interaction.values:
                
                coros.append(clicked(interaction))
                
            await asyncio.gather(*coros)
            
    async def __handleModalInteraction__(self, interaction: Interaction):
        
        target = None
        
        for modal in self.modals:
            
            if modal.customID == interaction.customID:
                
                target = modal
            
        if target:
            
            self.modals.remove(target)
            
            async def checkFilled():
                
                targetInputs: list[TextInput] = []
                
                nonlocal target, interaction
                raw = interaction.filled
                
                for filled in raw:
                    
                    comp: dict = filled.get('components')[0]
                    customID: int = comp.get('custom_id')
                    value: str = comp.get('value')
                    
                    for input in target.inputs:
                        
                        if input.customID == customID:
                            
                            targetInput = input
                            
                    if targetInput is not None:
                        
                        targetInput.value = value
                        targetInputs.append(targetInput)
                
                return targetInputs
            
            await interaction.ack()
            inputs = await checkFilled()
            
            coros = [target(interaction)]
            
            for input in inputs:
                
                coros.append(input(interaction, input.value))
                
            await asyncio.gather(*coros)
        
    async def __handlePresenceUpdate__(self, data: dict):
        
        user = data.pop('user')
        
        user = await self.bot.fetchUser(
            id = user['id']
        )
        
        presence = {
            'status': data['status'],
            'activities': data['activities'],
            'mobile': 'mobile' in data['client_status']
        }
        
        await asyncio.gather(*[event(user, presence) for event in self.presenceUpdateFunctions])
        return user, presence
        
    async def __handleChannelUpdate__(self, data):
        id = data['id']
        
        channel = await Channel.fromID(
            bot = self.bot,
            channel = id
        )
        
        for idx, cached in enumerate(self.channels):
            if cached.id == int(id):
                self.channels[idx] = Channel(self.bot, data)
                
        
        await asyncio.gather(*[event(channel) for event in self.channelUpdateFunctions])
        return channel
            
    async def __handleTyping__(self, data: dict):
        
        coros = [
            
            self.bot.fetchUser(
                id = data['user_id']
            ),
        
            self.bot.fetchChannel(
                id = data['channel_id']
            )
            
        ]
        
        user, channel = await asyncio.gather(*coros)
        
        self.typingTimes[(user.id, channel.id)] = time.time()
        
        await asyncio.gather(*[event(user, channel) for event in self.typingStartedFunctions])
        return user, channel
        
    async def __handleTypingStop__(self, user: User, channel: Channel):
        
        if False: # Disabled
            if user is not None and channel is not None:
            
                saved = self.typingTimes.get((user.id, channel.id))
                
                if saved is None:
                    return None
                
                gathered = [
                    event(user, channel)
                    for event in self.typingStoppedFunctions
                ]
                
                await asyncio.gather(*gathered)
                
            return user, channel
            
    async def __handleMessageUpdate__(self, data: dict):
        
        message = Message(self.bot, data)
        
        if message:
            
            channel = self.bot.getChannel(message.channel)
            
            if channel:
                
                old = channel.getMessage(message.id)
                
                if old:
                    
                    message.author, message.channel = old.author, old.channel
                    await message.__proc__()
                    
                    try:
                        
                        message.channel.messages.replace(old, message)
                        
                    except:
                        
                        ...
                    
                    await asyncio.gather(*[event() for event in self.messageUpdatedFunctions])
                    await self.__handleTypingStop__(message.author, message.channel)
                    
                    return message
        
    async def __handleMessageDelete__(self, data: dict):
        
        id = int(data.get('id'))
        channel = int(data.get('channel_id'))
        target = await self.bot.fetchMessage(channel, id)
        
        if target:
        
            await asyncio.gather(*[event(target) for event in self.messageDeletedFunctions])
                
            if target in target.channel.messages:
                target.channel.messages.remove(target)

            for view in self.views:

                if target.id in view.messageIDs:
                    
                    try:
                        view.messages.remove(target)
                    except:
                        view.messages.remove(target.id)

                    if len(view.messages) == 0:
                        
                        view.delete() # Delete from DB

                    view.save()
                
            return target
            
    async def __handleMessageReaction__(self, data: dict):
        
        userID = data.get('user_id')
        messageID = data.get('message_id')
        channelID = data.get('channel_id')
        emojiData = data.get('emoji')
        
        reaction = Reaction(
            bot = self.bot,
            user = userID,
            message = messageID,
            channel = channelID,
            emoji = emojiData
        )
        
        await reaction.__proc__()
        
        await asyncio.gather(*[event(reaction) for event in self.reactionAddFunctions])
        return reaction
    
    async def __handleReactionRemove__(self, data: dict):

        userID = data.get("user_id")
        messageID = data.get("message_id")
        channelID = data.get("channel_id")
        emoji = data.get("emoji")
        burst = data.get("burst")

        message = await self.bot.fetchMessage(
            channel = channelID,
            id = messageID
        )

        if message:

            ename = emoji.get("name")
            eid = emoji.get("id")
            found = None
            
            for reaction in message.reactions:

                if reaction.emojiName == ename or ( reaction.emojiID is not None and reaction.emojiID == eid ):

                    found = reaction

            if type(found) is Reaction:

                message.reactions.remove(found)
            
            await asyncio.gather(*[event(found) for event in self.reactionRemoveFunctions])
            
    async def __handlePinsUpdate__(self, data: dict):
        
        channel: int = int(data.get('channel_id'))
        
        try:
            guild: int = int(data.get('guild_id'))
            
        except:
            guild = None # DM or Group Chat
            
        # print('PinsUpdateData', data)
        
        # How to know which MESSAGE is pinned?
        
    async def __channelPinsAck__(self, data: dict):
        
        # print('ChannelPinsAck Received, Data:', data)
        
        version: int = data.get('version')
        timestamp: str = data.get('timestamp')
        channel: int = int(data.get('channel_id'))
        
        # print('Channel Pins Ack received for Channel', channel)
        
    async def __auditLogEntryCreate__(self, data: dict):
            
        try:
        
            user: int = int(data.get('user_id', 0))
            target: int = int(data.get('target_id', 0))
            options: dict = data.get('options', {})
            id: int = int(data.get('id', 0))
            actionType: int = data.get('action_type', 0)
            guild: int = int(data.get('guild_id', 0))

        except:

            ...
        
    async def __sessionsReplace__(self, data: list[dict]):
        
        sessions: int = len(data)
        objects: list[GatewaySession] = []
        
        for session in data:
            
            #status: str = session.get('status')
            #sessionID: str = session.get('session_id')
            object = GatewaySession(self, session)
            objects.append(object)
            
        debug.log(
            '%p Active Sessions: %p',
            sessions, objects
        )
        
    async def __newGuild__(self, data: dict):
        
        isLazy = data.get('lazy')
        isNew = not isLazy
        
        guild = await Guild.__create__(
            bot = self.bot,
            data = data
        )
        
        if isNew:
            
            ... # OnGuildCreateEvent
            
        return guild
    
    async def __handleGuildMemberUpdate__(self, data: dict):

        userData = data.pop("user")
        guildID = data.pop("guild_id")
        guild = await self.bot.fetchGuild(guildID)

        if guild:

            member = await Member.__create__(
                bot = self.bot,
                user = userData,
                data = data,
                guild = guild
            )

            await asyncio.gather(*[x(member) for x in self.memberUpdateFunctions])

            return member
        
    async def __handleChannelCreate__(self, data: dict):

        id = data.get("id")

        channel = await self.bot.fetchChannel(
            id = int(id)
        )

        channel.guild.channels.append(channel)
        await asyncio.gather(*[x(channel) for x in self.channelCreateFunctions])

    async def __handleChannelDelete__(self, data: dict):

        id = data.get("id")

        channel = self.bot.getChannel(
            id = int(id)
        )

        if channel:

            channel.guild.channels.remove(channel)
            await asyncio.gather(*[x(channel) for x in self.channelDeleteFunctions])

    async def __handleGuildMemberAdd__(self, data: dict):

        user = data.get("user")
        guild = data.get("guild_id")
        id = user.get("id")

        member = await self.bot.fetchMember(
            guild = guild,
            id = id
        )

        if member:
            
            member.guild.members.append(member)
            await asyncio.gather(*[x(member) for x in self.memberJoinFunctions])

    async def __handleGuildMemberRemove__(self, data: dict):

        user = data.get("user")
        guild = data.get("guild_id")
        id = user.get("id")

        member = self.bot.getMember(
            guild = guild,
            id = id
        )

        if member:
            
            try:
                member.guild.members.remove(member)
            except:
                ...

            await asyncio.gather(*[x(member) for x in self.memberLeaveFunctions])
    
    async def __ready__(self, sessionData: dict):

        self.isReady = True
        
        self.relationships = lunarlist(*[{'user': User(self.bot, relationship.get('user')), 'type': RELATIONSHIP_TYPES.get(relationship.get('type')), 'id': relationship.get('type')} for relationship in sessionData.get('relationships')])

        for relationship in self.relationships:
            await relationship.get("user").__proc__()
            self.users.append(relationship.get("user"))

        # DoLoad
        
        channels = await self.manager.loadChannels()
        
        for data in channels:
            
            channel = Channel(self.bot, data)
            await channel.__proc__()
                
        coros = []
                
        for channel in [channel for channel in self.channels if channel.recipients]:
            
            for recipient in channel.recipients:
                
                recipientID = recipient.get('id')
                
                coros.append(
                    self.bot.fetchUser(recipientID)
                )
                
        await asyncio.gather(*coros) # Fetch users to cache them
        
        try:
            guilds = await self.manager.loadGuilds()
        except:
            guilds = []

        coros = []
        
        for guild in guilds:

            async def fetch():

                try:
                    await Guild.fromID(
                        bot = self.bot,
                        guild = guild.get('id')
                    )
                
                except:
                    pass
            
            coros.append(fetch())
            
        if coros:
            
            await asyncio.gather(*coros) # Fetch guilds to cache them

        # DoLoad Finish

        self.session = GatewaySession(self, sessionData)

        await asyncio.gather(*[event(self.bot) for event in self.onStartFunctions])
            
        return self.bot
    
    async def __resumed__(self, data: dict):

        if not self.resumed:

            self.resumed = True

        trace = data.get("_trace")
        string = trace[0]
        data = json.loads(string)


from typing import Any
from os.path import isfile, split
from .__core__.requests import Manager
from .__core__.addons import Utils
import random

class Path: ...
class Url: ...

class Color:

    def __init__(self, decimal: int | str):

        self.__decimal__ = int(decimal)
    
    @classmethod
    def fromRGB(cls, red: int | str = 0, green: int | str = 0, blue: int | str = 0):

        if isinstance(red, tuple):

            if len(red) != 3:

                if len(red) == 0:
                    red = (0,)

                red, green, blue = (red[0], 0, 0)
            
            else:
                
                red, green, blue = red
        
        decimal = red * 65536 + green * 256 + blue
        return cls(decimal)
    
    def toRGB(self):
        value = self.__decimal__
        
        red = value // 65536
        green = (value % 65536) // 256
        blue = value % 256
        
        return red, green, blue
    
    @classmethod
    def red(cls):
        return cls(0xE74C3C)
    
    @classmethod
    def green(cls):
        return cls(0x2ECC71)
    
    @classmethod
    def blue(cls):
        return cls(0x3498DB)
    
    @classmethod
    def darkRed(cls):
        return cls(0x992D22)
    
    @classmethod
    def darkGreen(cls):
        return cls(0x1F8B4C)
    
    @classmethod
    def darkBlue(cls):
        return cls(0x206694)
    
    @classmethod
    def blurple(cls):
        return cls(0x5865F2)
    
    @classmethod
    def purple(cls):
        return cls(0x3498DB)
    
    @classmethod
    def yellow(cls):
        return cls(0xFEE75C)
    
    @classmethod
    def gray(cls):
        return cls(0x979C9F)
    
    @classmethod
    def white(cls):
        return cls(0xFFFFFF)
    
    @classmethod
    def black(cls):
        return cls(0x000000)
    
    @classmethod
    def fuchsia(cls):
        return cls(0xFF00FF)
    
    @classmethod
    def lime(cls):
        return cls(0x00FF00)
    
    @classmethod
    def lightRed(cls):
        return cls(0xFF6666)
    
    @classmethod
    def lightBlue(cls):
        return cls(0x6666FF)
    
    @classmethod
    def lightYellow(cls):
        return cls(0xFFFF66)
    
    @classmethod
    def darkYellow(cls):
        return cls(0x999900)
    
    @classmethod
    def lightGreen(cls):
        return cls(0x66FF66)
    
    @classmethod
    def cyan(cls):
        return cls(0x66FFFF)
    
    @classmethod
    def orange(cls):
        return cls(0xFFB84D)
    
    @classmethod
    def brown(cls):
        return cls(0xBC8F8F)
    
    @classmethod
    def random(cls):
        return cls.fromRGB(*(random.randint(0, 255) for x in range(3)))

class EmbedArray:
    
    def __init__(self, *embeds):
        self.container = []
        self.container.extend(embeds)
        
    def __add__(self, embed):
        if isinstance(embed, Embed):
            self.container.append(embed)
        elif isinstance(embed, EmbedArray):
            for x in embed.container:
                if isinstance(x, Embed):
                    self.container.append(x)
                    
        return self
    
    def __or__(self, embed):
        return self.__add__(embed)
    
    def __iter__(self):
        return self.container.__iter__()
    
    def __next__(self):
        return self.container.__next__()
    
    def __len__(self):
        return self.container.__len__()
    
    def __str__(self):
        return 'EmbedArray(' + self.container.__str__().removeprefix('[').removesuffix(']') + ')'
                    
    async def __json__(self):
        
        jsonified = []
        
        for embed in self:
            embedJson = await embed.__json__()
            jsonified.append(embedJson)
            
        return jsonified
    
class Embed:
    
    def __init__(self, title: str = None, description: str = None, footer: str = None, footerIcon: str = None, thumbnail: str = None, color: Color = None, image: str = None, timestamp: int | float | str = None):
        
        self.title = title
        self.description = description
        self.footer = footer
        self.footerIcon = footerIcon
        self.thumbnail = thumbnail
        self.color = color
        self.timestamp = timestamp
        self.image = image
        
    async def __json__(self):
        
        json = {'footer': {}, 'thumbnail': {}, 'image': {}}
        
        if self.__title__ is not None:
            json['title'] = self.__title__
            
        if self.__description__ is not None:
            json['description'] = self.__description__
            
        if self.__footer__ is not None:
            json['footer']['text'] = self.__footer__
            
        if self.__footericon__ is not None:
            json['footer']['icon_url'] = self.__footericon__
            
        if self.thumbnail is not None:
            json['thumbnail']['url'] = self.thumbnail
            
        if self.__color__ is not None:
            json['color'] = self.__color__.__decimal__

        if self.__timestamp__ is not None:
            json['timestamp'] = self.__timestamp__

        if self.__image__ is not None:
            json['image']['url'] = self.__image__
        
        return json
        
    @property
    def title(self):
        return self.__title__
    
    @title.setter
    def title(self, value):
        self.__title__ = str(value) if type(value) is not str and value is not None else value
        
    @title.deleter
    def title(self):
        self.__title__ = None
        
    @property
    def description(self):
        return self.__description__
    
    @description.setter
    def description(self, value):
        self.__description__ = str(value) if type(value) is not str and value is not None else value
        
    @description.deleter
    def description(self):
        self.__description__ = None
        
    @property
    def footer(self):
        return self.__footer__
    
    @footer.setter
    def footer(self, value):
        self.__footer__ = str(value) if type(value) is not str and value is not None else value
        
    @footer.deleter
    def footer(self):
        self.__footer__ = None
        
    @property
    def footerIcon(self):
        return self.__footericon__
    
    @footerIcon.setter
    def footerIcon(self, value):
        self.__footericon__ = str(value) if type(value) is not str and value is not None else value
        
    @footerIcon.deleter
    def footerIcon(self):
        self.__footericon__ = None
        
    @property
    def color(self):
        return self.__color__
    
    @color.setter
    def color(self, value):
        
        if hasattr(value, '__iter__') and len(value) == 3:
            value = Color.fromRGB(*value)
            
        self.__color__ = value
        
    @color.deleter
    def color(self):
        self.__color__ = None

    @property
    def timestamp(self):
        return self.__timestamp__
    
    @timestamp.setter
    def timestamp(self, new):
        self.__timestamp__: str = Utils.unixToIso(new) if type(new) in (int, float) else str(new) if new is not None else None

    @timestamp.deleter
    def timestamp(self):
        self.__timestamp__ = None

    @property
    def image(self):
        return self.__image__
    
    @image.setter
    def image(self, value):
        self.__image__ = str(value) if type(value) is not str and value is not None else value
        
    @image.deleter
    def image(self):
        self.__image__ = None

    @property
    def thumbnail(self):
        return self.__thumbnail__
    
    @thumbnail.setter
    def thumbnail(self, new):
        self.__thumbnail__ = str(new) if type(new) is not str and new is not None else new

    @thumbnail.deleter
    def thumbnail(self):
        self.__thumbnail__ = None
        
    def __add__(self, embed):
        
        if isinstance(embed, Embed):
            
            return EmbedArray(self, embed)
        
        elif isinstance(embed, EmbedArray):
            
            newArray = EmbedArray()
            newArray.container.append(self)
            newArray.container.extend(embed.container)
            return newArray
        
    def __or__(self, embed):
        
        self.__add__(embed)
        
    def __eq__(self, embed):
        
        if isinstance(embed, Embed):
            
            return (
                self.__title__ == embed.__title__
                and self.__description__ == embed.__description__
                and self.__footer__ == embed.__footer__
                and self.__footericon__ == embed.__footericon__
            )
        
        return False
    
    def __ne__(self, embed):
        return self.__eq__(embed) == False
    
    def __repr__(self):
        
        title = '' if self.title is None else f'title="{self.title}" '
        description = '' if self.description is None else f'description="{self.description}" '
        footer = '' if self.footer is None else f'footer="{self.footer}" '
        
        return ('<Embed ' + title + description + footer).strip() + '>'
    
    def __str__(self):
        
        for x in (
            self.__title__,
            self.__description__,
            self.__footer__
        ):
            
            if x is not None:
                return x
            
        return self.__repr__()
    
class Attachment:
    
    def __init__(self, bot, data: bytes, name: str):
        '''
        Construct a new `lunarcord.Attachment` from the given `data`.
        
        Parameters
        ----------
        data: `bytes`
            The attachment's data, in raw bytes.
            
        name: `str`
            The attachment name that will be visible on Discord.
        '''
        
        self.bot = bot

        self.__bytes = data
        self.name = str(name)
    
    @classmethod
    def fromBytes(cls, bot, data: bytes, name: str):
        '''
        Construct a new `lunarcord.Attachment` from the given `data`.
        
        Parameters
        ----------
        data: `bytes`
            The attachment's data, in raw bytes.
            
        name: `str`
            The attachment name that will be visible on Discord.
            
        Returns
        -------
        attachment: `lunarcord.Attachment`
            The newly created attachment.
        '''
        
        return cls(bot, data, name)
        
    @classmethod
    def fromPath(cls, bot, path: Path, name: str = None):
        '''
        Construct a new `lunarcord.Attachment` from the given `path`.
        
        Parameters
        ----------
        path: `Path`
            The path where the target file is located in.
            
        name: `str`
            The attachment name that will be visible on Discord. If not given, it is taken from the path.
            
        Returns
        -------
        attachment: `lunarcord.Attachment`
            The newly created attachment.
        '''
        
        path: str = path
        
        if isfile(path):
            with open(path, 'rb') as file:
                data = file.read()
                head, tail = split(path)
                
                if name is None:
                    name = tail
        else:
            data = b''
            
            if name is None:
                name = 'unknown.png'
            
        return cls(bot, data, name)
    
    @classmethod
    async def fromUrl(cls, bot, url: Url, name: str):
        '''
        A coroutine that constructs a new `lunarcord.Attachment` from the given `url`.
        
        Parameters
        ----------
        url: `Url`
            The link associated with the attachment to read from.
            
        name: `str`
            The attachment name that will be visible on Discord.
            
        Returns
        -------
        attachment: `lunarcord.Attachment`
            The newly created attachment.
        '''
        
        try:
            
            data = await bot.__gateway__.manager.get(url, customUrl=True, returns=bytes)
            
        except:
            data = b''
            
        return cls(bot, data, name)
    
    @classmethod
    def fromID(cls, bot, id: int, name: str):

        url = f"https://cdn.discordapp.com/attachments/{id}"
        return cls.fromUrl(bot, url, name)
    
    @classmethod
    def fromPayload(cls, bot, payload: dict):

        return Attachment.fromUrl(
            bot = bot,
            url = payload.get("url"),
            name = payload.get("filename")
        )
    
    @classmethod
    async def __convert__(cls, object: str, default = None, bot = None, source = None):

        from .basetypes import Interaction

        if type(source) is not Interaction:
            return default
        
        try:
            payload = source.attachments[object]
            return await cls.fromPayload(bot, payload)
        except:
            return default
    
    def save(self, name: str = ...):

        if name in (None, ...):
            name = self.__name

        with open(name, "wb") as file:
            file.write(self.__bytes)
    
    def __json__(self):
        
        json = {self.name: self.__bytes}
        return json
    
    def __repr__(self):

        return f"<Attachment name=\"{self.name}\">"
    
    def __str__(self):

        return self.name
    
    fromData = fromBytes
    fromFile = fromPath
    fromLink = fromUrl


import inspect, datetime, sys, json, sqlite3, os, asyncio, sys, re, random, time
from typing import TypeVar, Generic, Callable, Iterator

from importlib.util import spec_from_file_location, module_from_spec

from string import ascii_letters as asciiLetters, digits
lettersAndDigits = asciiLetters + digits

T = TypeVar('T')

        
class SlashOption:
    def __init__(self, name: str = ..., description: str = ..., required: bool = True, options: list = ..., default = None):
        
        '''
        Represents a Slash Command option. This should be used for building bot Slash Commands.
        Below is a simple example of how this class can be used:
        
        ```
        @bot.slashCommand(description="Repeats your message.")
        async def say(
            interaction: Interaction,
            message: str = SlashOption(name="message", description="The message to repeat", default="No message provided.")
        ):
        
            await interaction.send(message)
        ```
        
        The above code demonstrates how you can use this class along with the `slashCommand` decorator in order to create the basic, famous "say" / "echo" command.
        Of course, you can use this class however you like in order to create slash command options as you like.
        The following is the list of all parameters you can use while creating an object of this class, and their explanations...
        
        Parameters
        ----------
        name: `str`
            The argument's name on discord. By default, this will be the same as the argument name in code.
        
        description: `str`
            The description seen on discord. If not given, a placeholder description is used.
            
        required: `bool`
            Whether this argument should be required (True) or optional (False). This is by default True (required).
            
        default: `Any`
            The value to be given to you when the command user doesn't really provide any value (this can only happen with optional/not required arguments; Required ones can never be empty).
        '''
        
        if name is ...:
            name = None
            
        if description is ...:
            description = None
            
        if options is ...:
            options = None
        
        self.name = name
        self.description = description
        self.options = options
        self.required = required
        self.default = default
        
    def setName(self, new: str = ...):
        
        if new is not ...:
            self.name = new
            
        return self
            
    def setDescription(self, new: str = ...):
        
        if new is not ...:
            self.description = new
            
        return self
            
    def makeOptional(self):
        
        self.required = False
        return self
        
    def makeRequired(self):
        
        self.required = True
        return self
        
    def setDefault(self, default = ...):
        
        if default is not ...:
            self.default = default
            
        return self
            
    def setOptions(self, options: list = ...):
        
        if options is ...:
            options = None
            
        self.options = options
        
        return self

class Utils:
    
    def argsCount(function):
        
        return len(Utils.arguments(function))
    
    def slashOptions(function, *, Types):
        
        # OPTION TYPES
        # 3: STRING
        # 4: INTEGER
        # 5: BOOLEAN
        # 6: USER
        # 7: CHANNEL
        # 8: ROLE
        # 9: MENTIONABLE
        # 10: NUMBER
        # 11: ATTACHMENT
        
        options = []
        
        params = Utils.arguments(function)
        allparams = params.copy()
        interaction = params.pop(0)
        
        User, Channel, Role, Member, Mentionable, Attachment = Types
        
        for param in params:
            
            paramtype = param.get('type')
            
            if paramtype is str:
                ftype = 3
                
            elif paramtype is int:
                ftype = 4
                
            elif paramtype is bool:
                ftype = 5
                
            elif paramtype is float:
                ftype = 10
                
            elif paramtype is Channel:
                ftype = 7
                
            elif paramtype in (User, Member):
                ftype = 6

            elif paramtype is Role:
                ftype = 8

            elif paramtype is Attachment:
                ftype = 11
                
            else:
                ftype = 3
                
            name = param.get('name')
            hint = param.get('default')
            
            if type(hint) is SlashOption:
                description = hint.description if hint.description is not None else 'No description provided'
                required = hint.required
                defaultval = hint.default
            else:
                description = 'No description provided'
                required = True
                defaultval = hint
            
            option = {'name': name, 'description': description, 'type': ftype, 'required': required, 'default': None}
            options.append(option)
            
        return options, allparams
        
    def istype(object, *types):
        return isinstance(object, types)
    
    def isstr(object):
        return Utils.istype(object, str)
    
    def isint(object):
        return Utils.istype(object, int)
    
    def isfloat(object):
        return Utils.istype(object, float)
    
    def isStringOrInt(object):
        return Utils.isstr(object) or Utils.isint(object)

    def isIntOrFloat(object):
        return Utils.isint(object) or Utils.isfloat(object)
    
    def isStringOrFloat(object):
        return Utils.isstr(object) or Utils.isfloat(object)
    
    def isStringIntOrFloat(object):
        return Utils.isstr(object) or Utils.isint(object) or Utils.isint(object)
    
    def toJson(object):
        try:
            jsonified = object.__json__()
        except:
            jsonified = {'error': 'NotJsonifiable'}
            
        return jsonified
    
    def manager(bot):
        return bot.__gateway__.manager
    
    def unixToIso(timestamp: int | float):
        return datetime.datetime.utcfromtimestamp(float(timestamp)).isoformat()
    
    def isoToUnix(timestamp: str):
        return datetime.datetime.fromisoformat(str(timestamp)).timestamp()
    
    def isAwaitable(function):
        return inspect.iscoroutinefunction(function)
    
    def arguments(function) -> list[dict]:
        
        signature = inspect.signature(function)
        params = list(signature.parameters.values())
        values = []
        
        for idx, param in enumerate(params):
            
            appears = str(param)
            name = [part.strip() for part in appears.split(':')]
            
            if len(name) == 1:
                appears = name[0]
                
            elif len(name) == 2:
                typehint = name[1]
                typehint = typehint.split('.')[-1]
                appears = f'{name}: {typehint}'
            
            data = {
                'name': param.name,
                'default': param.default,
                'required': param.default is param.empty,
                'type': param.annotation,
                'index': idx,
                'empty': param.empty,
                'isEmpty': lambda x: x is param.empty,
                'appears': appears,
                'isArgs': appears.startswith('*') and not appears.startswith('**'),
                'isKwargs': appears.startswith('**')
            }
            
            values.append(data)
            
        return values
    
    async def convert(value, annotation, empty, bot, default, source = None):
        
        if value is None:
            return value
        
        try:
            typename = type(annotation).__name__
        except:
            typename = None
            
        annotations = []
            
        if typename == 'UnionType':
            
            if hasattr(annotation, '__args__'):
                annotations = list(annotation.__args__)
                
            else:
                annotations = [annotation]
                
        else:
            
            annotations = [annotation]
            
        unfixed = type(value)
            
        for annotation in annotations:
            
            if unfixed == annotation:
                continue
            
            if hasattr(annotation, '__convert__'):
                
                try:
                    converter = annotation.__convert__
                    converterArgs = Utils.arguments(converter)
                    converterLen = len(converterArgs)
                    converterSupplies = [value]
                    
                    if converterLen > 1:
                        converterSupplies.append(empty)
                        
                    if converterLen > 2:
                        converterSupplies.append(bot)

                    if converterLen > 3:
                        converterSupplies.append(source)
                        
                    if Utils.isAwaitable(converter):
                        value = await converter(*converterSupplies)
                        
                    else:
                        value = converter(*converterSupplies)
                    
                except:
                    
                    #import traceback
                    #traceback.print_exc()
                    ...
                    
                if not value == empty:
                    break
                
            elif annotation in (str, int, float, list, tuple, dict, set):
                
                try:
                    value = annotation(value)
                    
                except:
                    ...
                
        if value == empty:
            value = default
            
        return value
    
    async def execute(function, *args, bot = None, params: list[dict] = None, **kwargs):
        
        if bot is not None:
            kwargs['bot'] = bot
        
        if len(args) == 1:
            args = [args[0]]
            
        else:
            args = list(args)
        
        if params is None:
            required: list[dict] = Utils.arguments(function)
            
        else:
            required: list[dict] = params
            
        passing: list = []
        
        for param in required:
            
            name: str = param.get('name')
            isRequired: bool = param.get('required')
            annotation: type = param.get('type')
            default = param.get('default')
            idx: int = param.get('index')
            empty: type = param.get('empty')
            isArgs: bool = param.get('isArgs')
            
            if type(default) is SlashOption:
                
                default = default.default
            
            if isArgs:
                
                value = args[idx:]
                value = [await Utils.convert(x, annotation, empty, bot, x, args[0]) for x in value]
                passing.extend(value)
                
            else:
            
                try:
                    value = args[idx]
                
                except:
                    value = default if default is not empty else None
                    
                if value is None and name in kwargs:
                    
                    value = kwargs.get(name)
                
                if type(value) is str and value.strip() == '':
                    value = None
                    
                value = await Utils.convert(value, annotation, empty, bot, default, args[0])
                
                passing.append(value)
        
        if Utils.isAwaitable(function):
            await function(*passing)
            
        else:
            function(*passing)
            
    def removeFirst(amount: int, string: str):
        
        if type(amount) is str:
            amount = len(amount)
        
        string: list = list(string)
        
        for x in range(amount):
            
            string.pop(0)
            
        return str(''.join(string))
    
    def removeLast(amount: int, string: str):
        
        if type(amount) is str:
            amount = len(amount)
        
        string: list = list(string)
        
        for x in range(amount):
            
            string.pop()
            
        return str(''.join(string))
    
    def floatParts(number: float):
        
        '''
        Splits a `float` to two parts - integer and decimal.
        '''
        
        if type(number) is not float:
            number = float(number)
            
        parts = str(number).split('.')
        
        integer = 0
        decimal = 0
        
        if len(parts) >= 1:
            integer = int(parts[0])
            
        if len(parts) >= 2:
            decimal = int(parts[1])
            
        return integer, float(f'0.{decimal}')
    
    def floatDecimal(number: float) -> int:
        '''
        Splits a `float` to two parts and returns the decimal (second part).
        '''
        
        integer, decimal = Utils.floatParts(number)
        return decimal
    
    def disableLine(code: list[str], line: int) -> list[str]:
        
        '''
        Disable a specific line from the given lines of code.
        Return the results as a list of strings representing lines.
        '''
        
        try:
            value = code[line]
            
        except:
            return code
        
        new = '# ' + value
        
        # new is value but
        # in a comment so
        # that value is
        # not executed.
        
        code[line] = new # Replace old value with new.
        
        return code # Return the final code.
    
    def disableLineWithFunction(code: list[str], function: Callable | str):
        
        '''
        Disable line(s) containing the given function.
        Return the updated code as a list of strings, representing lines.
        '''
        
        ignoring: list[int] = []
        
        if type(function) is not str:
            
            try:
                function = function.__name__
                
            except:
                return code
            
        done = False
        
        while not done:
            
            index = None
        
            for idx, line in enumerate(code):
                if function + '(' in line:
                    index = idx
                    
            if index is None or index in ignoring:
                
                done = True
                continue
            
            code = Utils.disableLine(code, index)
            ignoring.append(index)
                
        return code
    
    def removeDuplicates(iterable: list):
        
        '''
        Return a copy of the given list with all the duplicate values removed.
        '''
        
        return list(dict.fromkeys(iterable))
    
    def module(name: str, path: str, locations: list = None, exec: bool = True):

        spec = spec_from_file_location(name, path, submodule_search_locations=locations)
        module = module_from_spec(spec)

        if exec:
            spec.loader.exec_module(module)

        return module
    
    def getImports(lines: list[str]):

        imports: list[str] = []

        patterns = [
            r'^\s*import\s+.+',
            r'^\s*from\s+\S+\s+import\s+.+'
        ]

        combined = '|'.join(patterns)

        for line in lines:
            if re.match(combined, line):
                imports.append(line.removesuffix("\n"))

        return imports
    
    def getImportsRaw(lines):
        
        return "\n".join(Utils.getImports(lines))
    
    def functionToString(function: Callable, imports: bool = True):

        source = inspect.getsource(function)

        if not imports:
            return source
        
        mpath = inspect.getfile(function)
        mname = mpath.removesuffix(".py")
        module = Utils.module(mname, mpath, exec=False)
        msource = inspect.getsourcelines(module)[0]
        imports = Utils.getImportsRaw(msource)

        source = source.split("\n")

        if len(source) >= 0 and source[0].startswith("@"):
            source.pop(0) # Remove decorator

        source = "\n".join(source)

        return imports + "\n\n" + source
    
    def stringToFunction(string: str) -> Callable | None:

        sections = string.split("\n\n")
        imports = sections.pop(0)
        code = "\n\n".join(sections)

        exec(imports)
        before = locals()
        exec(code)
        after = locals()

        pairs = list(after.items())
        pairs.reverse()
        pairs = tuple(pairs)
        
        for key, value in pairs:

            if callable(value):

                return value
            
    def ending(amount: int):

        if amount == 1:
            return ""
        
        return "s"
    
    def generateString(length: int = 14, exceptions: list[int] = []):

        while True:

            string = "".join([random.choice(lettersAndDigits) for x in range(length)])

            if string not in exceptions:
                return string
    
class _(Exception):
    def __init__(self, start: str, message: str):
        self.start = start
        self.message = message
        
        sys.tracebacklimit = 0
        super().__init__(self.msg)
        
    @property
    def msg(self):
        return self.start + ' ' + self.message
    
class Debugger:
    def __init__(self, hide: bool = False, method: int = 2):
        self.method: int = method
        self.hide: bool = hide
        self.last: str = None
        self.logs: list = []
    
    @property
    def hiding(self):
        return self.hide
    
    @hiding.setter
    def hiding(self, value: bool):
        if isinstance(value, bool) and value in (False, True):
            self.hide = value
        else:
            return None
        
    def disable(self):
        
        self.hiding = True
        
    def enable(self):
        
        self.hiding = False
        
    def __send__(self, message: str, force: bool = False):
        if not self.hide or force:
            if self.method == 1: # Print
                print('[LUNARCORD] ' + str(message))
                
            elif self.method == 2:
                raise _('[LUNARCORD]', message)
            
            else:
                return
            
            self.last: str = message
            self.logs.append(message)
            
    def filter(self, message: str, replacer: str = '%p', values: tuple = ()):
        for value in values:
            value = str(value)
            message = message.replace(replacer, value, 1)
            
        return message
            
    def log(self, message: str, *values, force: bool = False):
        if len(values) == 1:
            values = [values[0]]
        
        message = self.filter(
            message,
            '%p',
            values
        )
        
        self.__send__(message, force)
        
class Signal:
    
    def __init__(self, *types):
        
        self.__arguments__ = types
        self.__argsCount__ = len(types)
        self.__connected__ = []
        self.data = None
        
    def connect(self, *functions):
        try:
            if hasattr(functions, '__iter__'):
                for function in functions:
                    try:
                        if callable(function):
                            self.__connected__.append(function)
                    except:
                        pass
        except:
            pass
        
    async def call(self, *supplies, ignored: list[Callable] = []):
        
        functions = self.__connected__
        
        for function in functions:
            
            if function in ignored:
                
                continue
            
            try:
                signature = inspect.signature(function)
                arguments = list(signature.parameters.keys())
                argsCount = len(arguments)
                supplied = []
                
                for s in supplies:
                    index = supplies.index(s)
                    if index + 1 <= argsCount:
                        signalArg = self.__arguments__[index]
                        if signalArg != type(s) and signalArg != object:
                            
                            try:
                                s = signalArg(s)
                                
                            except:
                                s = None
                                
                        supplied.append(s)
            except:
                continue
        
            unsupplied = argsCount - len(supplied)
            if unsupplied >= 1:
                for x in range(unsupplied):
                    supplied.append(None)
                    
            coro = function(*supplied)
            
            if Utils.isAwaitable(function):
                await coro
                
    def callSync(self, *supplies):
        '''
        Same as `Signal.call`, but ignores all `async def` functions so that you can also use this in a non-async context.
        '''
        
        functions = self.__connected__
        ignoring = [function for function in functions if Utils.isAwaitable(function)]
        
        for function in functions:
            
            if function in ignoring:
                
                continue
            
            try:
                signature = inspect.signature(function)
                arguments = list(signature.parameters.keys())
                argsCount = len(arguments)
                supplied = []
                
                for s in supplies:
                    index = supplies.index(s)
                    if index + 1 <= argsCount:
                        signalArg = self.__arguments__[index]
                        if signalArg != type(s) and signalArg != object:
                            
                            try:
                                s = signalArg(s)
                                
                            except:
                                s = None
                                
                        supplied.append(s)
            except:
                continue
        
            unsupplied = argsCount - len(supplied)
            if unsupplied >= 1:
                for x in range(unsupplied):
                    supplied.append(None)
                    
            function(*supplied)
            
    async def wait(self, loop: asyncio.AbstractEventLoop, timeout: float = None) -> tuple | None:
        
        '''
        Blocks your code until this signal has been called from anywhere.
        '''
        
        future = loop.create_future()
        
        def onComplete(*results):
            
            nonlocal future
            future.set_result(results)
            
        self.connect(onComplete)
        
        try:
            return await future

        finally:
            self.remove(onComplete)
        
        
    def disconnect(self, *functions):
        try:
            for function in functions:
                try:
                    self.__connected__.remove(function)
                except:
                    pass
        except:
            pass
        
    def updateArg(self, index, newType):
        try:
            self.__arguments__[index] = newType
        except:
            pass
        
    def argsList(self):
        return list(self.__arguments__)
    
    def connList(self):
        return list(self.__connected__)
        
    def removeArg(self, index):
        try:
            argList = self.argsList()
            del argList[index]
            argTuple = tuple(argList)
            self.__arguments__ = argTuple
            self.__argsCount__ -= 1
        except:
            pass
        
    def addArg(self, index, arg):
        try:
            argList = self.argsList()
            argList.insert(index, arg)
            argTuple = tuple(argList)
            self.__arguments__ = argTuple
            self.__argsCount__ += 1
        except:
            pass
        
    def appendArg(self, arg):
        try:
            self.addArg(self.__argsCount__, arg)
        except:
            pass
        
    def resetConns(self):
        for conn in self.conns():
            self.disconnect(conn)
            
    async def __call__(self, *supplies):
        
        return await self.call(*supplies)

    async def emit(self, *supplies): return await self.call(*supplies)
    def emitSync(self, *supplies): return self.callSync(*supplies)
    def add(self, *functions): return self.connect(*functions)
    def remove(self, *functions): return self.disconnect(*functions)
    def args(self): return self.argsList()
    def conns(self): return self.connList()
    def add(self, index, arg): return self.addArg(index, arg)
    def append(self, arg): return self.appendArg(arg)
    def reset(self): return self.resetConns()

class ComboSignal:

    """A signal that connects multiple signals altogether."""

    def __init__(self, *signals: Signal):

        setattr(self, "__container__", [])
        self.__container__.extend(signals)

    def __iter__(self) -> Iterator[Signal]:

        return self.__container__.__iter__()
    
    def connect(self, *callbacks: Callable):

        for signal in self:
            signal.connect(*callbacks)

    async def emit(self, *supplies):

        coros = []

        for signal in self:

            coros.append(
                signal.emit(*supplies)
            )

        await asyncio.gather(*coros)

    def emitSync(self, *supplies):

        for signal in self:
            signal.emitSync(*supplies)
    
class DatabaseHandler:
    
    def __init__(self, path: str):
        
        self.path: str = path
        self.tables: list = []
        self.phs: list = []
        
        self.connect(path=path)
        
    def connect(self, path: str):
        
        '''
        Create a new DataBase connection to the given `path`.
        '''
        
        self.connection = sqlite3.connect(path, check_same_thread=False)
        self.cursor = self.connection.cursor()
        
    def close(self):
        
        '''
        Close the current DataBase connection.
        '''
        
        if self.connection:
            
            self.connection.close()
            self.connection = None
            
    def reconnect(self):
        
        '''
        Close and re-create a connection.
        '''
        
        self.close()
        
        self.connect(
            path = self.path
        )
        
    def createTable(self, name: str, types: dict[str, type]):
        
        '''
        Creates a new table with the given `name` and `types`.
        '''
        
        try:
        
            values = [f'{x} {self.typeToStr(types.get(x))}' for x in types]
            values = ', '.join(values + ['Data json'])
            self.cursor.execute(f'CREATE TABLE {name} ({values})')
            
        except:
            
            ...
            
        self.phs.append(f'{list(types.keys())[0]}, DATA')
        self.tables.append(name)
            
        return Table(self, name, types)
        
    def typeToStr(self, annotation: type):
        
        '''
        Converts `annotation` type to a string, eg `int` to `INTEGER`
        '''
        
        
        if annotation is int:
            return 'INTEGER'
        
        if annotation is str:
            return 'TEXT'
        
        return None
    
    def insertData(self, key: str, data, table: str = None):
        
        '''
        Insert a new `key` with data `data` for table `table`
        '''
        
        if table is None:
            
            try:
                table = self.tables[0]
                
            except:
                return None
            
        index = self.tables.index(table)
        ph = self.phs[index]
        
        string = self.toJson(data, ascii=False)
        
        self.execute(
            f'INSERT INTO {table} ({ph}) VALUES (?, ?)',
            key, string
        )
        
        self.commit() # Update all changes
        
    def loadData(self, key: str, table: str = None):
        
        '''
        Load data for `key` in table `table`
        '''
        
        if table is None:
            
            try:
                table = self.tables[0]
                
            except:
                return None
            
        index = self.tables.index(table)
        ph = self.phs[index]
        auth: str = ph.split(',')[0].strip()
        
        self.execute(
            f'SELECT * from {table} where {auth}=?',
            key
        )
        
        loaded = self.cursor.fetchall()
        return self.fromJson(loaded[0][1])
    
    def loadAll(self, table: str = None):
        
        '''
        Load all `key`s
        '''
        
        all = []
        
        if table is None:
            
            try:
                table = self.tables[0]
                
            except:
                return None
        
        self.execute(
            f'SELECT * from {table}'
        )
        
        data = self.cursor.fetchall()
        
        for loaded in data:
            
            all.append(
                self.fromJson(loaded[1])
            )
            
        return all
    
    def delete(self, key: str, table: str):
        
        '''
        Delete `key` from table of name `table`
        '''
        
        if table is None:
            
            try:
                table = self.tables[0]
                
            except:
                return None
            
        ph = self.phs[
            self.tables.index(table)
        ].split(',')[0].strip()
        
        self.execute(
            f'DELETE from {table} where {ph}=?',
            key
        )
        
        self.commit()
        
    def update(self, key: str, item: str, value, table: str = None):
        
        '''
        Replace the value for key `item` with the new `value`, where the target is `key` in `table`.
        '''
        
        if table is None:
            
            try:
                table = self.tables[0]
                
            except:
                return None
            
        phs = self.phs[
            self.tables.index(table)
        ].split(',')
        
        auth = phs[0].strip()
        data = phs[1].strip()
        
        current = self.loadData(key, table)
        current[item] = value
        
        self.execute(
            f'UPDATE {table} SET {data}=? where {auth}=?',
            self.toJson(current, ascii=False), key
        )
        
        self.commit()
        
        
    def toJson(self, data: dict | list, ascii: bool = True) -> str:
        
        '''
        Convert data to JavaScript Object Notation (JSON) string.
        '''
        
        return json.dumps(data, ensure_ascii=ascii)
    
    def fromJson(self, data: str) -> dict | list:
        
        '''
        Convert JavaScript Object Notation (JSON) string to pythonic data.
        '''
        
        return json.loads(data)
    
    def commit(self):
        
        '''
        Commit all changes to the connection.
        '''
        
        self.connection.commit()
        
    def execute(self, action: str, *params):
        
        '''
        Execute the given sqlite3 `action` with the additional `params`.
        '''
        
        if len(params) == 1:
            params = (params[0],)
        
        self.cursor.execute(action, params)
        
class Table:
    
    def __init__(self, db: DatabaseHandler, name: str, types: dict[str, type]):
        
        '''
        Represents a database table.
        '''
        
        self.db = db
        self.name = name
        self.types = types
        
    def insert(self, key: str, data = {}):
        
        '''
        Insert data to the table.
        '''
        
        self.db.insertData(
            key, data, self.name
        )
        
    def delete(self, key: str):
        
        '''
        Delete data of key `key`.
        '''
        
        self.db.delete(
            key, self.name
        )
        
    def update(self, key: str, item: str, value):
        
        '''
        Update `item` from data of key `key` with new `value`.
        '''
        
        self.db.update(
            key, item, value, self.name
        )
        
    def load(self, key: str):
        
        '''
        Load data for key `key`.
        '''
        
        return self.db.loadData(
            key, self.name
        )
        
    def loadAll(self):
        
        '''
        Load data for all keys.
        '''
        
        return self.db.loadAll(
            table = self.name
        )
    
    def exists(self, key: str):

        '''
        Check if key `key` really exists.
        '''

        try:
            self.load(key=key)
            return True
        except:
            return False
        
    def add(self, key: str, data = {}):

        '''
        Checks if `key` already exists. If not, inserts the `data`. Otherwise, ignores the operation.
        '''

        if not self.exists(key):
            self.insert(key, data)
        
class Database:
    
    def __init__(
        self, origin: str, path: str = None
    ):
        
        if path is None:
            path = 'lunardata.db'
            
        full: str = path
        
        if origin is not None:
            full: str = os.path.join(origin, path)
        
        self.manager = DatabaseHandler(full)
        
        self.channels = self.manager.createTable(
            'channels', {'id': int}
        )
        
        self.users = self.manager.createTable(
            'users', {'id': int}
        )

        self.views = self.manager.createTable(
            'views', {'name': str}
        )

        self.timers = self.manager.createTable(
            'timers', {'type': str}
        )

        for x in range(1):
            
            self.timers.add("loops", {})
            self.timers.add("cooldowns", {})
        
    def createChannel(self, id: int):
        
        self.channels.insert(
            key = id
        )
        
    def createUser(self, id: int):
        
        self.users.insert(
            key = id
        )
        
    def updateChannel(self, id: int, item: str, value):
        
        self.channels.update(
            key = id,
            item = item,
            value = value
        )
        
    def updateUser(self, id: int, item: str, value):
        
        self.users.update(
            key = id,
            item = item,
            value = value
        )
        
    def loadChannel(self, id: int):
        
        '''
        Load all saved data for a channel and return it or `None` if it is not found in the database
        '''
        
        try:
            return self.channels.load(
                key = id
            )
            
        except:
            return None
        
    def loadUser(self, id: int):
        
        '''
        Load all saved data for a user and return it or `None` if it is not found in the database
        '''
        
        try:
            return self.users.load(
                key = id
            )
            
        except:
            return None
        
    def loadChannels(self):
        
        return self.channels.loadAll()
    
    def loadUsers(self):
        
        return self.users.loadAll()
            
    def loadView(self, name: str):

        try:
            return self.views.load(
                key = name
            )
        
        except:
            return None
        
    def loadViews(self):

        return self.views.loadAll()
    
    def viewExists(self, name):

        return self.loadView(name) is not None
    
    def createView(self, name: str):

        self.views.insert(
            key = name
        )

    def deleteView(self, name: str):

        self.views.delete(
            key = name
        )
    
    def updateView(self, name, item, value):

        self.views.update(
            key = name,
            item = item,
            value = value
        )

    def loadLoops(self) -> dict:
        return self.timers.load("loops")
    
    def loadCooldowns(self) -> dict:
        return self.timers.load("cooldowns")
    
    def createLoop(self, name: str, timer: int | float):
        self.timers.update("loops", item=name, value=timer)

    def createCooldown(self, name: str, timer: int | float):
        self.timers.update("cooldowns", item=name, value=timer)

    def loadLoop(self, name: str) -> int | float:

        loops = self.loadLoops()

        for lname, timer in loops.items():
            if lname == name:
                return timer
            
    def loadCooldown(self, name: str) -> int | float:

        cds = self.loadCooldowns()

        for cdname, timer in cds.items():
            if cdname == name:
                return timer
    
class lunarlist(Generic[T], list):
    
    @property
    def length(self):
        '''
        The list's `length`. This is the same as doing `len(myList)` where `myList` is a `lunarlist`.
        '''
        
        return self.__len__()
    
    @property
    def list(self):
        '''
        A `list` equivalent of this `lunarlist`.
        '''
        
        return self.toList()
    
    @property
    def tuple(self):
        '''
        A `tuple` equivalent of this `lunarlist`.
        '''
        
        return self.toTuple()
    
    @property
    def set(self):
        '''
        A `set` equivalent of this `lunarlist`.
        '''
        
        return self.toSet()
    
    @property
    def reversed(self):
        '''
        A `reversed` copy of this `lunarlist`.
        '''
        
        copy = self.copy()
        copy.reverse()
        return copy
    
    def __init__(self, *arguments):
        
        '''
        A mutable, list-like sequence with additional methods and attributes.
        '''
        
        if len(arguments) == 1:
            arguments = [arguments[0]]
            
        else:
            arguments = list(arguments)
        
        super().__init__(arguments)
    
    def replace(self, old: T, new: T):
        
        '''
        Removes `old` value and adds `new` at its index.
        If `old` does not exist, a `ValueError` will be raised.
        '''
        
        try:
            idx = self.index(old)
        except:
            raise ValueError(f"'{old}' not in list")
        
        self.pop(idx)
        self.insert(idx, new)
        
    def toList(self) -> list:
        
        '''
        Converts this `lunarlist` to a built-in Python `list`.
        '''
        
        return list(self)
    
    def toTuple(self) -> tuple:
        
        '''
        Converts this `lunarlist` to a built-in Python `tuple`.
        '''
        
        return tuple(self)
    
    def toSet(self) -> set:
        
        '''
        Converts this `lunarlist` to a built-in Python `set`.
        '''
        
        return set(self)
    
    def convert(self, __type: type):
        
        '''
        Converts this `lunarlist` to an object of type `__type`.
        '''
        
        if __type is list:
            return self.toList()
        
        if __type is tuple:
            return self.toTuple()
        
        if __type is set:
            return self.toSet()
        
        if type(__type) is not type:
            raise TypeError(f'Invalid __type argument')
        
        raise ValueError(f'Cannot convert to \'{__type.__str__}\'')
    
    def appendSafe(self, value: T):
        
        '''
        Appends a value to the list, without worrying about if it already is in.
        '''
        
        super().append(value)
        
    def append(self, value: T):
        
        '''
        Append a value to the list, if it is not in it already.
        '''
        
        if value not in self:
            self.appendSafe(value)
            
    def reverseGet(self, value, default = None):
        
        '''
        Doing a reverse `lunarlist.get()` search, returns the index of the given object.
        '''
        
        for idx, x in enumerate(self):
            
            if x == value:
                
                return idx
            
        return default
    
    def __iter__(self) -> Iterator[T]:

        return super().__iter__()
    
class Time:
    
    def __init__(self, seconds: int | float = 0, minutes: int | float = 0, hours: int | float = 0, days: int | float = 0, weeks: int | float = 0, months: int | float = 0, years: int | float = 0):
        
        self.seconds: int | float = seconds
        
        self.seconds += Time.minutesToSeconds(minutes)
        self.seconds += Time.hoursToSeconds(hours)
        self.seconds += Time.daysToSeconds(days)
        self.seconds += Time.weeksToSeconds(weeks)
        self.seconds += Time.monthsToSeconds(months)
        self.seconds += Time.yearsToSeconds(years)
        
    @property
    def minutes(self):
        '''
        Converts this `Time` object to a `float` representing its time in minutes.
        '''
        
        return Time.secondsToMinutes(self.seconds)
    
    @property
    def hours(self):
        '''
        Converts this `Time` object to a `float` representing its time in hours.
        '''
        
        return Time.secondsToHours(self.seconds)
    
    @property
    def days(self):
        '''
        Converts this `Time` object to a `float` representing its time in days.
        '''
        
        return Time.secondsToDays(self.seconds)
    
    @property
    def weeks(self):
        '''
        Converts this `Time` object to a `float` representing its time in weeks.
        '''
        
        return Time.secondsToWeeks(self.seconds)
    
    @property
    def months(self):
        '''
        Converts this `Time` object to a `float` representing its time in months.
        '''
        
        return Time.secondsToMonths(self.seconds)
    
    @property
    def years(self):
        '''
        Converts this `Time` object to a `float` representing its time in years.
        '''
        
        return Time.secondsToHours(self.years)
        
    def __int__(self):
        return int(self.seconds)
    
    def __float__(self):
        return float(self.seconds)
    
    def __str__(self):
        return str(int(self))
    
    def format(self, formatter: str):
        
        minutes, seconds = Time.secondsToMinutesAndSeconds(self.seconds)
        
        return formatter.replace(
            '%s', str(seconds)
        ).replace(
            '%m', str(minutes)
        )
        
    @classmethod
    def minutesToSeconds(cls, minutes: int | float):
        
        return minutes * 60
    
    @classmethod
    def secondsToMinutes(cls, seconds: int | float):
        
        return seconds / 60
    
    @classmethod
    def hoursToSeconds(cls, hours: int | float):
        
        return hours * 60 * 60
    
    @classmethod
    def secondsToHours(cls, seconds: int | float):
        
        return seconds / 60 / 60
    
    @classmethod
    def daysToSeconds(cls, days: int | float):
        
        return days * 60 * 60 * 24
    
    @classmethod
    def secondsToDays(cls, seconds: int | float):
        
        return seconds / 60 / 60 / 24
    
    @classmethod
    def weeksToSeconds(cls, weeks: int | float):
        
        return weeks * 60 * 60 * 24 * 7
    
    @classmethod
    def secondsToWeeks(cls, seconds: int | float):
        
        return seconds / 60 / 60 / 24 / 7
    
    @classmethod
    def monthsToSeconds(cls, months: int | float):
        
        return months * 60 * 60 * 24 * 30
    
    @classmethod
    def secondsToMonths(cls, seconds: int | float):
        
        return seconds / 60 / 60 / 24 / 30
    
    @classmethod
    def yearsToSeconds(cls, years: int | float):
        
        return years * 60 * 60 * 24 * 365
    
    @classmethod
    def secondsToYears(cls, seconds: int | float):
        
        return seconds / 60 / 60 / 24 / 365
    
    @classmethod
    def minutesToHours(cls, minutes: int | float):
        
        return cls.secondsToHours(cls.minutesToSeconds(minutes))
    
    @classmethod
    def minutesToDays(cls, minutes: int | float):
        
        return cls.secondsToDays(cls.minutesToSeconds(minutes))
    
    @classmethod
    def minutesToWeeks(cls, minutes: int | float):
        
        return cls.secondsToWeeks(cls.minutesToSeconds(minutes))
    
    @classmethod
    def minutesToMonths(cls, minutes: int | float):
        
        return cls.secondsToMonths(cls.minutesToSeconds(minutes))
    
    @classmethod
    def minutesToYears(cls, minutes: int | float):
        
        return cls.secondsToYears(cls.minutesToSeconds(minutes))
    
    @classmethod
    def hoursToDays(cls, hours: int | float):
        
        return cls.secondsToDays(cls.hoursToSeconds(hours))
    
    @classmethod
    def hoursToMinutes(cls, hours: int | float):
        
        return cls.secondsToMinutes(cls.hoursToSeconds(hours)) 
    
    @classmethod
    def hoursToWeeks(cls, hours: int | float):
        
        return cls.secondsToWeeks(cls.hoursToSeconds(hours))
    
    @classmethod
    def hoursToMonths(cls, hours: int | float):
        
        return cls.secondsToMonths(cls.hoursToSeconds(hours))
    
    @classmethod
    def hoursToYears(cls, hours: int | float):
        
        return cls.secondsToYears(cls.hoursToSeconds(hours))

    @classmethod
    def daysToMinutes(cls, days: int | float):
        
        return cls.secondsToMinutes(cls.daysToSeconds(days))
        
    @classmethod
    def daysToHours(cls, days: int | float):
        
        return cls.secondsToHours(cls.daysToSeconds(days))
    
    @classmethod
    def daysToWeeks(cls, days: int | float):
        
        return cls.secondsToWeeks(cls.daysToSeconds(days))
    
    @classmethod
    def daysToMonths(cls, days: int | float):
        
        return cls.secondsToMonths(cls.daysToSeconds(days))
    
    @classmethod
    def daysToYears(cls, days: int | float):
        
        return cls.secondsToYears(cls.daysToSeconds(days))
    
    @classmethod
    def weeksToHours(cls, weeks: int | float):
        
        return cls.secondsToHours(cls.weeksToSeconds(weeks))
    
    @classmethod
    def weeksToMinutes(cls, weeks: int | float):
        
        return cls.secondsToMinutes(cls.weeksToSeconds(weeks))
    
    @classmethod
    def weeksToDays(cls, weeks: int | float):
        
        return cls.secondsToDays(cls.weeksToSeconds(weeks))
    
    @classmethod
    def weeksToMonths(cls, weeks: int | float):
        
        return cls.secondsToMonths(cls.weeksToSeconds(weeks))
    
    @classmethod
    def weeksToYears(cls, weeks: int | float):
        
        return cls.secondsToYears(cls.weeksToSeconds(weeks))
    
    @classmethod
    def monthsToHours(cls, months: int | float):
        
        return cls.secondsToMonths(cls.monthsToSeconds(months))
    
    @classmethod
    def monthsToMinutes(cls, months: int | float):
        
        return cls.secondsToMinutes(cls.monthsToSeconds(months))
    
    @classmethod
    def monthsToDays(cls, months: int | float):
        
        return cls.secondsToDays(cls.monthsToSeconds(months))
    
    @classmethod
    def monthsToWeeks(cls, months: int | float):
        
        return cls.secondsToWeeks(cls.monthsToSeconds(months))
    
    @classmethod
    def monthsToYears(cls, months: int | float):
        
        return cls.secondsToYears(cls.monthsToSeconds(months))
    
    @classmethod
    def yearsToHours(cls, years: int | float):
        
        return cls.secondsToHours(cls.yearsToSeconds(years))
    
    @classmethod
    def yearsToMinutes(cls, years: int | float):
        
        return cls.secondsToMinutes(cls.yearsToSeconds(years))
    
    @classmethod
    def yearsToDays(cls, years: int | float):
        
        return cls.secondsToDays(cls.yearsToSeconds(years))
    
    @classmethod
    def yearsToWeeks(cls, years: int | float):
        
        return cls.secondsToWeeks(cls.yearsToSeconds(years))
    
    @classmethod
    def yearsToMonths(cls, years: int | float):
        
        return cls.secondsToMonths(cls.yearsToSeconds(years))
    
    @classmethod
    def secondsToMinutesAndSeconds(cls, seconds: int | float) -> tuple[int, int]:
        
        minutes = float(cls.secondsToMinutes(seconds))
        integer, decimal = Utils.floatParts(minutes)
        
        if decimal > 0:
            decimal = cls.minutesToSeconds(decimal)
            
        return integer, round(decimal)
    
    @classmethod
    def secondsToHoursMinutesSeconds(cls, seconds: int | float) -> tuple[int, int]:
        
        minutes, seconds = cls.secondsToMinutesAndSeconds(seconds)
        integer, decimal = Utils.floatParts(float(cls.minutesToHours(minutes)))
        
        if decimal > 0:
            decimal = cls.hoursToMinutes(decimal)
            
        return integer, round(decimal), seconds
        
class Cooldown:

    """An object representing data for a command's cooldown."""

    def __init__(self, payload: dict[str]):

        self.ends: int | float = payload.get("ends")

    @classmethod
    def construct(cls, ends: int | float):

        """For internal use only. Constructs a new `Cooldown` object."""

        current = time.time()

        payload = {
            "ends": ends,
            "remains": ends - current,
            "active": ends is not None and current < ends
        }

        return cls(payload)
    
    @property
    def remaining(self) -> float:

        """The remaining time until the cooldown ends."""

        try:
            return self.ends - time.time()
        
        except:
            return 0.0
    
    @property
    def active(self):
        """Whether this cooldown is currently active."""
        return self.remaining > 0
    
    @property
    def timestamp(self):
        """A relative timestamp to use on Discord."""
        return f"<t:{int(self.ends)}:R>"
    
    def __str__(self):
        return self.timestamp
    
    def __repr__(self):
        return str(self)
    
    def __float__(self):
        return self.ends
    
    def __int__(self):
        return int(float(self))
    
    def __bool__(self):
        return self.active
    
'''time = Time(minutes=1, seconds=30)
print(time.secondsToHoursMinutesSeconds(3700))'''